#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function

import os
import sys
import subprocess
import logging
import shutil
import stat

logger = logging.getLogger('wasienv')

# Part of this implementation is taken/inspired from Emscripten tools/shared.py
# https://github.com/emscripten-core/emscripten/blob/2431347a32dcce89ab3e26e86de445cada58745c/tools/shared.py#L140-L191
# LICENSE below:

# Copyright (c) 2010-2014 Emscripten authors, see AUTHORS file.
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

class Py2CalledProcessError(subprocess.CalledProcessError):
  def __init__(self, returncode, cmd, output=None, stderr=None):
    super(Exception, self).__init__(returncode, cmd, output, stderr)
    self.returncode = returncode
    self.cmd = cmd
    self.output = output
    self.stderr = stderr


class Py2CompletedProcess:
  def __init__(self, args, returncode, stdout, stderr):
    self.args = args
    self.returncode = returncode
    self.stdout = stdout
    self.stderr = stderr

  def check(self):
    if self.returncode != 0:
      raise Py2CalledProcessError(returncode=self.returncode, cmd=self.args, output=self.stdout, stderr=self.stderr)


def check_program(cmd):
    if not os.path.exists(cmd):
        raise Exception("The program {} was not found. Is the SDK installed?\nYou can install it via: wasienv install-sdk unstable".format(os.path.basename(cmd)))


def python2_subprocess_run(cmd, check=True, input=None, *args, **kwargs):
  if input is not None:
    kwargs['stdin'] = subprocess.PIPE

  proc = subprocess.Popen(cmd, *args, **kwargs)
  stdout, stderr = proc.communicate(input)
  result = Py2CompletedProcess(cmd, proc.returncode, stdout, stderr)
  if check:
    result.check()
  return result


def run_process(cmd, check=True, input=None, *args, **kwargs):
  # print("wasienv run process: {}".format(" ".join(cmd)))
  logger.debug("wasienv run process: {}".format(" ".join(cmd)))
  debug_text = '%sexecuted %s' % ('successfully ' if check else '', ' '.join(cmd))

  run = getattr(subprocess, "run", python2_subprocess_run)
  ret = run(cmd, check=check, input=input, *args, **kwargs)
  logger.debug(debug_text)

  return ret.returncode


def is_exe(fpath):
    return os.path.isfile(fpath) and os.access(fpath, os.X_OK)


def is_wasm(fpath):
    with open(fpath, 'rb') as f:
        first_bytes = f.read(4)
        # is_wasm = first_bytes == [0x00, 0x61, 0x73, 0x6d]
        is_wasm = bytearray(first_bytes) == b'\x00asm'
        f.seek(0)
        return is_wasm
    return False


def try_to_wrap_executable(exe_name):
    target_path = os.path.join(os.getcwd(), exe_name)
    if not is_exe(target_path) or exe_name.endswith(".wasm") or exe_name.endswith(".dylib") or exe_name.endswith(".dll") or exe_name.endswith(".so") or exe_name.endswith(".a"):
        return
    # It's a cmake file, we skip
    # CMake does some checks like the size of a struct generating
    # a file with certain contents on it and then doing a check using
    # regex. That means that we can't wrap it in a executable
    if os.path.basename(exe_name).startswith('cmTC_'):
        return

    st = os.stat(target_path)
    if not is_wasm(target_path):
        return

    new_target_path = "{}.wasm".format(target_path)
    shutil.copy(target_path, new_target_path)
    with open(target_path,'w') as f:
        f.write("#!/bin/bash\nwasirun {} \"$@\"\n".format(new_target_path))


    os.chmod(new_target_path, st.st_mode)

    # Copy files to the temp folder 
    # Wasm file
    temp_target_path = os.path.join('/tmp', os.path.split(new_target_path)[1])
    shutil.copy(new_target_path, temp_target_path)
    # Executable file
    temp_target_path = os.path.join('/tmp', os.path.split(target_path)[1])
    shutil.copy(target_path, temp_target_path)


def find_output_arg(args):
  """Find and remove any -o arguments.  The final one takes precedence.
  Return the final -o target along with the remaining (non-o) arguments.
  """
  outargs = []
  specified_target = None
  use_next = False
  for arg in args:
    if use_next:
      specified_target = arg
      use_next = False
      continue
    if arg == '-o':
      use_next = True
    elif arg.startswith('-o'):
      specified_target = arg[2:]
    else:
      outargs.append(arg)
  return specified_target, outargs


def set_environ():
    from .constants import WASI_CC, WASI_CXX, WASI_LD, WASI_AR, WASI_RANLIB, WASI_NM

    os.environ["CC"] = WASI_CC
    os.environ["CXX"] = WASI_CXX
    os.environ["LD"] = WASI_LD
    os.environ["AR"] = WASI_AR
    os.environ["RANLIB"] = WASI_RANLIB
    os.environ["NM"] = WASI_NM

    os.environ["WASI_CC"] = WASI_CC
    os.environ["WASI_CXX"] = WASI_CXX
    os.environ["WASI_LD"] = WASI_LD
    os.environ["WASI_AR"] = WASI_AR
    os.environ["WASI_RANLIB"] = WASI_RANLIB
    os.environ["WASI_NM"] = WASI_NM


def execute(f):
    try:
        logging.basicConfig(level=logging.INFO)
        sys.exit(f(sys.argv))
    except KeyboardInterrupt:
        logger.warning("KeyboardInterrupt")
        sys.exit(1)



import os

from .sdk import WASI_SDK_DIR

wasienv_DIR    = os.path.dirname(__file__)

# The location of the underlying binaries
CC      = os.path.join(WASI_SDK_DIR, "bin/clang")
LD      = os.path.join(WASI_SDK_DIR, "bin/wasm-ld")
CXX     = os.path.join(WASI_SDK_DIR, "bin/clang++")
NM      = os.path.join(WASI_SDK_DIR, "bin/llvm-nm")
AR      = os.path.join(WASI_SDK_DIR, "bin/llvm-ar")
RANLIB  = os.path.join(WASI_SDK_DIR, "bin/llvm-ranlib")

WASI_SYSROOT = os.path.join(WASI_SDK_DIR, "share/wasi-sysroot")

# Binaries
WASI_CC      = "wasicc"
WASI_CXX     = "wasic++"
WASI_LD      = "wasild"
WASI_AR      = "wasiar"
WASI_NM      = "wasinm"
WASI_RANLIB  = "wasiranlib"
WASI_RUN  = "wasirun"

WASI_CMAKE   = os.path.abspath(os.path.join(wasienv_DIR, "wasienv.cmake"))

STUBS_SYSTEM_LIB = os.path.join(wasienv_DIR, "stubs")
STUBS_SYSTEM_PREAMBLE = os.path.join(STUBS_SYSTEM_LIB, "preamble.h")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function

from .tools import logger, run_process, execute, check_program
from .constants import AR


def run(args):
    check_program(AR)
    proc_args = [AR]+args[1:]
    return_code = run_process(proc_args, check=False)
    return return_code


if __name__ == '__main__':
    execute(run)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function

import sys

from .tools import logger, run_process, find_output_arg, try_to_wrap_executable, execute, check_program
from .constants import LD, WASI_SYSROOT



def run(args):
    # has_sysroot = any([arg.startswith("--sysroot") for arg in args])
    # has_target = any([arg.startswith("--target") for arg in args])

    # if not has_sysroot:
    #     args.append("--sysroot={}".format(WASI_SYSROOT))
    
    # if not has_target:
    #     args.append("--target=wasm32-wasi")
    check_program(LD)
    proc_args = [LD]+args[1:]
    return_code = run_process(proc_args, check=False)
    target, outargs = find_output_arg(args)
    args.append('-lwasi-emulated-mman')
    if target:
        try_to_wrap_executable(target)
    return return_code


if __name__ == '__main__':
    execute(run)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function

import sys

from .tools import logger, run_process, execute



def run(args):
    if len(args) < 2:
        # It should be wasirun x.wasm
        raise Exception("You need to provide a WebAssembly file")
    filename = args[1]
    proc_args = ["wasmer", "run", "--dir=.", "--enable-all", filename, "--"] + args[2:]
    return_code = run_process(proc_args, check=False)
    return return_code


if __name__ == '__main__':
    execute(run)



import os
import sys
import platform
try:
    # Python 3
    from io import BytesIO as StringIO
except ImportError:
    from StringIO import StringIO

import requests
import tarfile

from .tools import logger

# Where the python packages live
PACKAGES_DIR    = os.path.dirname(os.path.dirname(__file__))

# Where the wasienv storage lives
WASI_STORAGE_DIR    = os.path.join(PACKAGES_DIR, "wasienv-storage")
WASI_SDKS_DIR = os.path.join(WASI_STORAGE_DIR, "sdks")

CURRENT_SDK = "7"

SDKS = {
    "5": {
        "download_urls": {
            "darwin": 'https://github.com/CraneStation/wasi-sdk/releases/download/wasi-sdk-5/wasi-sdk-5.0-macos.tar.gz',
            "linux": 'https://github.com/CraneStation/wasi-sdk/releases/download/wasi-sdk-5/wasi-sdk-5.0-linux.tar.gz',
        },
        "sysroot": "wasi-sdk-5.0/opt/wasi-sdk"
    },
    "6": {
        "download_urls": {
            "darwin": 'https://github.com/CraneStation/wasi-sdk/releases/download/wasi-sdk-6/wasi-sdk-6.0-macos.tar.gz',
            "linux": 'https://github.com/CraneStation/wasi-sdk/releases/download/wasi-sdk-6/wasi-sdk-6.0-linux.tar.gz',
        },
        "sysroot": "wasi-sdk-6.0/opt/wasi-sdk"
    },
    "7": {
        "download_urls": {
            "darwin": 'https://github.com/CraneStation/wasi-sdk/releases/download/wasi-sdk-7/wasi-sdk-7.0-macos.tar.gz',
            "linux": 'https://github.com/CraneStation/wasi-sdk/releases/download/wasi-sdk-7/wasi-sdk-7.0-linux.tar.gz',
        },
        "sysroot": "wasi-sdk-7.0/opt/wasi-sdk"
    },
}

SDK_TAGS = {
    "latest": "7",
    "unstable": "7"
}


class SDKException(Exception):
    pass


class SDKAlreadyExists(SDKException):
    pass


class SDKNotInstalled(SDKException):
    pass


def get_sdk_dir(sdk_name):
    return os.path.join(WASI_SDKS_DIR, sdk_name)


def unalias_name(sdk_name):
    if sdk_name in SDK_TAGS:
        sdk_name = SDK_TAGS[sdk_name]
    return sdk_name


def get_sdk(sdk_name):
    sdk = SDKS.get(sdk_name)
    if not sdk:
        raise SDKException("The SDK {} doesn't exist".format(sdk_name))
    return sdk


def is_sdk_installed(sdk_name):
    sdk_name = unalias_name(sdk_name)
    sdk = get_sdk(sdk_name)
    sdk_dir = get_sdk_dir(sdk_name)
    return os.path.exists(sdk_dir) and os.path.isdir(sdk_dir) and bool(os.listdir(sdk_dir))


def download_and_unpack(sdk_name):
    sdk_name = unalias_name(sdk_name)
    sdk = get_sdk(sdk_name)
    download_urls = sdk.get("download_urls", {})
    if is_sdk_installed(sdk_name):
        raise SDKAlreadyExists("The SDK {} is already installed".format(sdk_name))
    system = platform.system()
    download_url = download_urls.get(system.lower(), None)
    if not download_url:
        raise SDKException("The SDK {} doesn't have an installable for {}".format(sdk_name, system))

    sdk_dir = get_sdk_dir(sdk_name)
    logger.info("Downloading WASI SDK to: {} from: {}".format(sdk_dir, download_url))
    response = requests.get(download_url)
    logger.info("SDK Downloaded, uncompressing")
    tf = tarfile.open(fileobj=StringIO(response.content))
    os.makedirs(sdk_dir)
    tf.extractall(sdk_dir)
    # Assert the sysroot dir was created
    sysroot = get_sdk_sysroot(sdk, sdk_name)
    if not os.path.isdir(sysroot):
        raise SDKException("The SDK expected sysroot doesn't exist: {}".format(sysroot))
    logger.info("SDK installed successfully")


def get_sdk_sysroot(sdk, sdk_name):
    if not sdk:
        raise Exception("SDK {} not found".format(sdk_name))
    if "sysroot" not in sdk:
        raise Exception("SDK {} doesn't have a sysroot defined".format(sdk_name))
    return os.path.join(get_sdk_dir(sdk_name), sdk["sysroot"])


def set_default_sdk(sdk_name):
    sdk_name = unalias_name(sdk_name)
    sdk = get_sdk(sdk_name)
    if not is_sdk_installed(sdk_name):
        raise SDKNotInstalled("The SDK {} is not installed".format(sdk_name))


WASI_SDK = get_sdk(CURRENT_SDK)
WASI_SDK_DIR = get_sdk_sysroot(WASI_SDK, CURRENT_SDK)


if __name__ == '__main__':
  try:
    sys.exit(download_and_unpack(CURRENT_SDK))
  except KeyboardInterrupt:
    logger.warning("KeyboardInterrupt")
    sys.exit(1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function

import sys
import os

from .tools import logger, run_process, set_environ, execute
from .constants import WASI_CMAKE, WASI_SDK_DIR



def run(args):
    # If using cmake, we point to the toolchain file directly
    set_environ()

    if len(args) <= 1:
        print("You need to run wasimake with make or cmake (eg. `wasimake cmake .`)")
        exit(1)

    if args[1] == "cmake":
        os.environ["WASI_SDK_DIR"] = WASI_SDK_DIR
        proc_args = ["cmake", "-DCMAKE_TOOLCHAIN_FILE={}".format(WASI_CMAKE)]+args[2:]
        return_code = run_process(proc_args, check=False)
        return return_code

    return_code = run_process(args[1:], check=False)
    return return_code


if __name__ == '__main__':
    execute(run)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function

import sys
import os
from .tools import execute
from .sdk import download_and_unpack, CURRENT_SDK, SDKAlreadyExists, set_default_sdk


def run(args):
    if len(args) <= 1:
        print("wasienv command line tool")
        return
    if args[1] == "install-sdk":
        sdk_version = args[2]
        try:
            download_and_unpack(sdk_version)
        except SDKAlreadyExists as e:
            print(e)
    elif args[1] == "default-sdk":
        if len(args) == 3:
            sdk_version = args[2]
            set_default_sdk(sdk_version)
        else:
            print(CURRENT_SDK)


if __name__ == '__main__':
    execute(run)



from .tools import execute
from .wasiar import run as wasiar_run
from .wasicc import run as wasicc_run
from .wasmcc import run as wasmcc_run
from .wasiconfigure import run as wasiconfigure_run
from .wasienv import run as wasienv_run
from .wasild import run as wasild_run
from .wasimake import run as wasimake_run
from .wasinm import run as wasinm_run
from .wasiranlib import run as wasiranlib_run
from .wasirun import run as wasirun_run


def wasiar():
    return execute(wasiar_run)


def wasicc():
    return execute(wasicc_run)

def wasmcc():
    return execute(wasmcc_run)

def wasiconfigure():
    return execute(wasiconfigure_run)


def wasienv():
    return execute(wasienv_run)


def wasild():
    return execute(wasild_run)


def wasimake():
    return execute(wasimake_run)


def wasinm():
    return execute(wasinm_run)


def wasiranlib():
    return execute(wasiranlib_run)


def wasirun():
    return execute(wasirun_run)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function

import sys
import os

from .tools import logger, run_process, set_environ, execute
from .constants import WASI_CMAKE, WASI_SDK_DIR


def run(args):
    set_environ()

    if len(args) <= 1:
        print("You need to run wasiconfigure with another command (eg. `wasiconfigure ./configure`)")
        exit(1)

    return_code = run_process(args[1:], check=False)
    return return_code


if __name__ == '__main__':
    execute(run)



# -*- coding: utf-8 -*-

from setuptools import setup
import os
import codecs


CURRENT_DIR = os.path.dirname(__file__)


def get_long_description():
    readme_md = os.path.join(CURRENT_DIR, "README.md")
    with open(readme_md) as ld_file:
        return ld_file.read()


setup(
    name='wasi',
    description="wasi is a C/C++ Compiler toolchain for WASI",
    # long_description=get_long_description(),
    long_description_content_type="text/markdown",
    keywords="webassembly wasi wasienv wasmer",
    author="Syrus Akbary",
    author_email="syrus@wasmer.io",
    url="https://github.com/wasienv-core/wasienv",
    version='0.4.1',
    packages=['wasienv'],
    include_package_data=True,
    license="MIT",
    zip_safe=False,
    install_requires=[
        "requests",
    ],
    entry_points={'console_scripts': [
        'wasiar = wasienv.commands:wasiar',
        'wasienv = wasienv.commands:wasienv',
        'wasicc = wasienv.commands:wasicc',
        'wasic++ = wasienv.commands:wasicc',
        'wasmcc = wasienv.commands:wasmcc',
        'wasmc++ = wasienv.commands:wasmcc',
        'wasiconfigure = wasienv.commands:wasiconfigure',
        'wasild = wasienv.commands:wasild',
        'wasimake = wasienv.commands:wasimake',
        'wasirun = wasienv.commands:wasirun',
        'wasinm = wasienv.commands:wasinm',
        'wasiranlib = wasienv.commands:wasiranlib',
    ]},
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function

import sys
import os

from .tools import logger, run_process, try_to_wrap_executable, find_output_arg, execute, check_program
from .constants import CC, CXX, WASI_SYSROOT, STUBS_SYSTEM_LIB, STUBS_SYSTEM_PREAMBLE


def run(args):
    main_program = CXX if args[0].endswith("wasmc++") else CC
    check_program(main_program)
    if '--version' in args:
        print('''wasienv (wasienv gcc/clang-like replacement)''')
        return 0

    if len(args) == 1 and args[0] == '-v': # -v with no inputs
        # autoconf likes to see 'GNU' in the output to enable shared object support
        print('wasienv (wasienv gcc/clang-like replacement + linker emulating GNU ld)', file=sys.stderr)
        code = run_process([main_program, '-v'], check=False).returncode
        return code

    has_target = any([arg.startswith("--target") for arg in args])

    # Flags decided by following: https://github.com/wasienv/wasienv/pull/8
    args.append("--no-standard-libraries")
    args.append("-Wl,--export-all")
    args.append("-Wl,--no-entry")

    if not has_target:
        args.append("--target=wasm32-unknown-unknown")

    proc_args = [main_program]+args[1:]
    return_code = run_process(proc_args, check=False)
    target, outargs = find_output_arg(args)
    if target:
        try_to_wrap_executable(target)
    return return_code


if __name__ == '__main__':
    execute(run)
